from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_phase(
    app_with_lock_datetime_in_past: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_lock_datetime_in_past)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_matches_db"),
    )
    initialize_database(engine_for_test, app_with_lock_datetime_in_past)

    # Signup one random user
    user_name = get_random_string(6)
    first_name = get_random_string(10)
    last_name = get_random_string(8)
    password = get_random_string(15)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": """
                mutation Root(
                    $userName: String!, $firstName: String!,
                    $lastName: String!, $password: String!
                ) {
                    signupResult(
                        userName: $userName, firstName: $firstName,
                        lastName: $lastName, password: $password
                    ) {
                        __typename
                        ... on UserWithToken {
                            token
                        }
                        ... on UserNameAlreadyExists {
                            message
                        }
                    }
                }
            """,
            "variables": {
                "userName": user_name,
                "firstName": first_name,
                "lastName": last_name,
                "password": password,
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"

    token = response_signup.json()["data"]["signupResult"]["token"]

    # Check retrieve all phases
    query_all_phases = """
        query {
            allPhasesResult {
                __typename
                ... on Phases {
                    phases {
                        id
                        code
                        description
                        binaryBets {
                            team1 {
                                description
                                won
                            }
                            team2 {
                                description
                                won
                            }
                        }
                        scoreBets {
                            team1 {
                                description
                                score
                            }
                            team2 {
                                description
                                score
                            }
                        }
                        groups {
                            id
                            description
                            code
                            scoreBets {
                                id
                                locked
                                team1 {
                                    score
                                    description
                                }
                                team2 {
                                    score
                                    description
                                }
                            }
                            binaryBets {
                                id
                                locked
                                team1 {
                                    won
                                    description
                                }
                                team2 {
                                    won
                                    description
                                }
                            }
                        }
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
            }
        }
    """

    response_all_phases = client.post(
        "/api/v2", headers={"Authorization": f"Bearer {token}"}, json={"query": query_all_phases}
    )

    assert response_all_phases.json() == {
        "data": {
            "allPhasesResult": {
                "__typename": "Phases",
                "phases": [
                    {
                        "id": ANY,
                        "code": "GROUP",
                        "description": "Phase de groupes",
                        "binaryBets": [],
                        "scoreBets": [
                            {
                                "team1": {"description": "Andorre", "score": None},
                                "team2": {"description": "Brésil", "score": None},
                            },
                            {
                                "team1": {"description": "Burkina Faso", "score": None},
                                "team2": {
                                    "description": "Guatemala",
                                    "score": None,
                                },
                            },
                            {
                                "team1": {"description": "Andorre", "score": None},
                                "team2": {"description": "Burkina Faso", "score": None},
                            },
                            {
                                "team1": {"description": "Brésil", "score": None},
                                "team2": {
                                    "description": "Guatemala",
                                    "score": None,
                                },
                            },
                            {
                                "team1": {"description": "Andorre", "score": None},
                                "team2": {
                                    "description": "Guatemala",
                                    "score": None,
                                },
                            },
                            {
                                "team1": {"description": "Brésil", "score": None},
                                "team2": {"description": "Burkina Faso", "score": None},
                            },
                        ],
                        "groups": [
                            {
                                "binaryBets": [],
                                "code": "A",
                                "description": "Groupe A",
                                "id": ANY,
                                "scoreBets": [
                                    {
                                        "id": ANY,
                                        "locked": True,
                                        "team1": {"description": "Andorre", "score": None},
                                        "team2": {"description": "Brésil", "score": None},
                                    },
                                    {
                                        "id": ANY,
                                        "locked": True,
                                        "team1": {"description": "Burkina Faso", "score": None},
                                        "team2": {
                                            "description": "Guatemala",
                                            "score": None,
                                        },
                                    },
                                    {
                                        "id": ANY,
                                        "locked": True,
                                        "team1": {"description": "Andorre", "score": None},
                                        "team2": {"description": "Burkina Faso", "score": None},
                                    },
                                    {
                                        "id": ANY,
                                        "locked": True,
                                        "team1": {"description": "Brésil", "score": None},
                                        "team2": {
                                            "description": "Guatemala",
                                            "score": None,
                                        },
                                    },
                                    {
                                        "id": ANY,
                                        "locked": True,
                                        "team1": {"description": "Andorre", "score": None},
                                        "team2": {
                                            "description": "Guatemala",
                                            "score": None,
                                        },
                                    },
                                    {
                                        "id": ANY,
                                        "locked": True,
                                        "team1": {"description": "Brésil", "score": None},
                                        "team2": {"description": "Burkina Faso", "score": None},
                                    },
                                ],
                            },
                        ],
                    },
                ],
            },
        },
    }

    # Error case : authentication error
    response_error_auth_all_phases = client.post("/api/v2", json={"query": query_all_phases})

    assert response_error_auth_all_phases.json() == {
        "data": {
            "allPhasesResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }

    # Success case : Retrieve phase by id
    query_phase_by_id = """query Root($phaseId: UUID!) {
        phaseByIdResult(id: $phaseId) {
            __typename
            ... on Phase {
                id
                code
                description
            }
            ... on InvalidToken {
                message
            }
            ... on ExpiredToken {
                message
            }
            ... on PhaseByIdNotFound {
                message
            }
        }
    }
    """

    phase_id = response_all_phases.json()["data"]["allPhasesResult"]["phases"][0]["id"]

    response_phase_by_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": query_phase_by_id,
            "variables": {"phaseId": phase_id},
        },
    )

    assert response_phase_by_id.json() == {
        "data": {
            "phaseByIdResult": {
                "__typename": "Phase",
                "id": phase_id,
                "code": "GROUP",
                "description": "Phase de groupes",
            },
        },
    }

    # Error case : phase with invalid id
    invalid_phase_id = str(uuid4())

    response_phase_with_invalid_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": query_phase_by_id,
            "variables": {"phaseId": invalid_phase_id},
        },
    )

    assert response_phase_with_invalid_id.json() == {
        "data": {
            "phaseByIdResult": {
                "__typename": "PhaseByIdNotFound",
                "message": f"Phase not found: {invalid_phase_id}",
            },
        },
    }

    # Error case : authentication error
    response_authentication_error = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {get_random_string(5)}"},
        json={
            "query": query_phase_by_id,
            "variables": {"phaseId": phase_id},
        },
    )

    assert response_authentication_error.json() == {
        "data": {
            "phaseByIdResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }
    # Success case : Fetch phase by code
    query_phase_by_code = """query Root($phaseCode: String!) {
        phaseByCodeResult(code: $phaseCode) {
            __typename
            ... on Phase {
                id
                code
                description
            }
            ... on InvalidToken {
                message
            }
            ... on ExpiredToken {
                message
            }
            ... on PhaseByCodeNotFound {
                message
            }
        }
    }
    """

    phase_code = response_phase_by_id.json()["data"]["phaseByIdResult"]["code"]

    response_phase_by_code = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={"query": query_phase_by_code, "variables": {"phaseCode": phase_code}},
    )

    assert response_phase_by_code.json() == {
        "data": {
            "phaseByCodeResult": {
                "__typename": "Phase",
                "id": phase_id,
                "code": phase_code,
                "description": "Phase de groupes",
            },
        },
    }

    # Error case : Phase with invalid code
    invalid_phase_code = get_random_string(6)

    response_phase_with_invalid_code = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={"query": query_phase_by_code, "variables": {"phaseCode": invalid_phase_code}},
    )

    assert response_phase_with_invalid_code.json() == {
        "data": {
            "phaseByCodeResult": {
                "__typename": "PhaseByCodeNotFound",
                "message": f"Phase not found: {invalid_phase_code}",
            },
        },
    }

    # Error case : authentication error
    response_authentication_error = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {get_random_string(5)}"},
        json={"query": query_phase_by_code, "variables": {"phaseCode": phase_code}},
    )

    assert response_authentication_error.json() == {
        "data": {
            "phaseByCodeResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }



from http import HTTPStatus
from typing import TYPE_CHECKING

from starlette.testclient import TestClient

if TYPE_CHECKING:
    from fastapi import FastAPI


def test_endpoint_not_found(app_with_valid_jwt_config: "FastAPI") -> None:
    client = TestClient(app_with_valid_jwt_config)

    response = client.get("/api/v1/fffffff")

    assert response.status_code == HTTPStatus.NOT_FOUND
    assert response.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": "Not Found",
    }



from http import HTTPStatus
from typing import TYPE_CHECKING

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_group(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_valid_jwt_config)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_language"),
    )
    initialize_database(engine_for_test, app_with_valid_jwt_config)

    # Signup one random user
    user_name = get_random_string(6)
    first_name = get_random_string(10)
    last_name = get_random_string(8)
    password = get_random_string(9)

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": user_name,
            "first_name": first_name,
            "last_name": last_name,
            "password": password,
        },
    )

    auth_token = response_signup.json()["result"]["token"]

    # retrieve all bets with default language
    response_default_lang = client.get(
        "/api/v1/bets",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_default_lang.json()["result"]["phases"][0]["description"] == "Phase de groupes"
    assert response_default_lang.json()["result"]["groups"][0]["description"] == "Groupe A"
    assert response_default_lang.json()["result"]["score_bets"][0]["team1"]["description"] == "Laos"
    assert (
        response_default_lang.json()["result"]["score_bets"][0]["team2"]["description"]
        == "Thaïlande"
    )

    # retrieve all bets with french language
    response_french_lang = client.get(
        "/api/v1/bets?lang=fr",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert (
        response_default_lang.json()["result"]["phases"][0]["description"]
        == response_french_lang.json()["result"]["phases"][0]["description"]
    )
    assert (
        response_default_lang.json()["result"]["groups"][0]["description"]
        == response_french_lang.json()["result"]["groups"][0]["description"]
    )
    assert (
        response_default_lang.json()["result"]["score_bets"][0]["team1"]["description"]
        == response_french_lang.json()["result"]["score_bets"][0]["team1"]["description"]
    )
    assert (
        response_default_lang.json()["result"]["score_bets"][0]["team1"]["description"]
        == response_french_lang.json()["result"]["score_bets"][0]["team1"]["description"]
    )

    # retrieve all bets with english language
    response_english_lang = client.get(
        "/api/v1/bets?lang=en",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_english_lang.json()["result"]["phases"][0]["description"] == "Group stage"
    assert response_english_lang.json()["result"]["groups"][0]["description"] == "Group A"
    assert response_english_lang.json()["result"]["score_bets"][0]["team1"]["description"] == "Laos"
    assert (
        response_english_lang.json()["result"]["score_bets"][0]["team2"]["description"]
        == "Thailand"
    )

    # retrieve all bets with unknown language
    response_german_lang = client.get(
        "/api/v1/bets?lang=de",
        headers={"Authorization": f"Bearer {auth_token}"},
    )

    assert response_german_lang.status_code == HTTPStatus.UNPROCESSABLE_ENTITY
    assert response_german_lang.json() == {
        "description": [
            {
                "ctx": {"expected": "'fr' or 'en'"},
                "type": "enum",
                "input": "de",
                "loc": ["query", "lang"],
                "msg": "Input should be 'fr' or 'en'",
            },
        ],
        "error_code": HTTPStatus.UNPROCESSABLE_ENTITY,
        "ok": False,
    }



from secrets import choice
from typing import TYPE_CHECKING
from uuid import uuid4

import pendulum
from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database
from yak_server.helpers.settings import get_settings

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_binary_bet(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_valid_jwt_config)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_binary_bet"),
    )
    initialize_database(engine_for_test, app_with_valid_jwt_config)

    user_name = get_random_string(10)
    password = get_random_string(30)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": """
                mutation Root(
                    $userName: String!, $firstName: String!,
                    $lastName: String!, $password: String!
                ) {
                    signupResult(
                        userName: $userName, firstName: $firstName,
                        lastName: $lastName, password: $password
                    ) {
                        __typename
                        ... on UserWithToken {
                            token
                        }
                        ... on UserNameAlreadyExists {
                            message
                        }
                    }
                }
            """,
            "variables": {
                "userName": user_name,
                "firstName": get_random_string(6),
                "lastName": get_random_string(12),
                "password": password,
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"

    response_login = client.post(
        "/api/v2",
        json={
            "query": """
                mutation Root($userName: String!, $password: String!) {
                    loginResult(userName: $userName, password: $password) {
                        __typename
                        ... on UserWithToken {
                            token
                            binaryBets {
                                id
                            }
                        }
                        ... on InvalidCredentials {
                            message
                        }
                    }
                }
            """,
            "variables": {
                "userName": user_name,
                "password": password,
            },
        },
    )

    assert response_login.json()["data"]["loginResult"]["__typename"] == "UserWithToken"
    token = response_login.json()["data"]["loginResult"]["token"]
    bet_id = response_login.json()["data"]["loginResult"]["binaryBets"][0]["id"]

    mutation_modify_binary_bet = """
        mutation Root($id: UUID!, $isOneWon: Boolean) {
            modifyBinaryBetResult(id: $id, isOneWon: $isOneWon) {
                __typename
                ... on BinaryBet {
                    id
                    group {
                        description
                        phase {
                            description
                        }
                    }
                    team1 {
                        won
                        description
                    }
                    team2 {
                        won
                        description
                    }
                }
                ... on BinaryBetNotFoundForUpdate {
                    message
                }
                ... on LockedBinaryBetError {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on InvalidToken {
                    message
                }
            }
        }
    """

    # Success case : modify binary bet
    is_one_won = choice([True, False, None])

    response_modify_binary_bet = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": mutation_modify_binary_bet,
            "variables": {"id": bet_id, "isOneWon": is_one_won},
        },
    )

    assert response_modify_binary_bet.json() == {
        "data": {
            "modifyBinaryBetResult": {
                "__typename": "BinaryBet",
                "group": {"description": "Groupe A", "phase": {"description": "Phase de groupes"}},
                "id": bet_id,
                "team1": {
                    "description": "France",
                    "won": None if is_one_won is None else is_one_won,
                },
                "team2": {
                    "description": "Allemagne",
                    "won": None if is_one_won is None else not is_one_won,
                },
            },
        },
    }

    # Error case : authentication error

    response_modify_binary_bet_error_auth = client.post(
        "/api/v2",
        json={
            "query": mutation_modify_binary_bet,
            "variables": {"id": bet_id, "isOneWon": is_one_won},
        },
    )

    assert response_modify_binary_bet_error_auth.json() == {
        "data": {
            "modifyBinaryBetResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }

    # Error case : Modify with invalid id
    invalid_bet_id = str(uuid4())

    response_modify_binary_bet_with_invalid_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": mutation_modify_binary_bet,
            "variables": {"id": invalid_bet_id, "isOneWon": True},
        },
    )

    assert response_modify_binary_bet_with_invalid_id.json() == {
        "data": {
            "modifyBinaryBetResult": {
                "__typename": "BinaryBetNotFoundForUpdate",
                "message": "Binary bet not found. Cannot modify a resource that does not exist.",
            },
        },
    }

    # Error case : locked binary bet
    app_with_valid_jwt_config.dependency_overrides[get_settings]().set_lock_datetime(
        -pendulum.duration(seconds=10)
    )

    response_modify_locked_binary_bet = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": mutation_modify_binary_bet,
            "variables": {"id": bet_id, "isOneWon": True},
        },
    )

    assert response_modify_locked_binary_bet.json() == {
        "data": {
            "modifyBinaryBetResult": {
                "__typename": "LockedBinaryBetError",
                "message": "Cannot modify binary bet, lock date is exceeded",
            },
        },
    }

    app_with_valid_jwt_config.dependency_overrides[get_settings]().set_lock_datetime(
        pendulum.duration(seconds=10)
    )

    # Success case : Retrieve one binary bet
    query_binary_bet = """
        query getBinaryBet($id: UUID!) {
            binaryBetResult(id: $id) {
                __typename
                ... on BinaryBet {
                    id
                    group {
                        description
                        phase {
                            description
                        }
                    }
                    team1 {
                        won
                        description
                    }
                    team2 {
                        won
                        description
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on BinaryBetNotFound {
                    message
                }
            }
        }
    """

    response_retrieve_binary_bet = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={"query": query_binary_bet, "variables": {"id": bet_id}},
    )

    assert response_retrieve_binary_bet.json() == {
        "data": {
            "binaryBetResult": {
                "__typename": "BinaryBet",
                "group": {"description": "Groupe A", "phase": {"description": "Phase de groupes"}},
                "id": bet_id,
                "team1": {
                    "description": "France",
                    "won": None if is_one_won is None else is_one_won,
                },
                "team2": {
                    "description": "Allemagne",
                    "won": None if is_one_won is None else not is_one_won,
                },
            },
        },
    }

    # Error case : retrieve binary bet with invalid id
    invalid_bet_id = str(uuid4())

    response_binary_with_invalid_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {token}"},
        json={"query": query_binary_bet, "variables": {"id": invalid_bet_id}},
    )

    assert response_binary_with_invalid_id.json() == {
        "data": {
            "binaryBetResult": {
                "__typename": "BinaryBetNotFound",
                "message": f"Binary bet not found: {invalid_bet_id}",
            },
        },
    }

    # Error case : authentication error
    response_retrieve_binary_bet = client.post(
        "/api/v2", json={"query": query_binary_bet, "variables": {"id": bet_id}}
    )

    assert response_retrieve_binary_bet.json() == {
        "data": {
            "binaryBetResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            },
        },
    }



import logging
from typing import Annotated

import pendulum
from fastapi import APIRouter, Depends, status
from pydantic import UUID4
from sqlalchemy.orm import Session

from yak_server.database.models import UserModel
from yak_server.helpers.authentication import (
    NameAlreadyExistsError,
    encode_bearer_token,
    signup_user,
)
from yak_server.helpers.database import get_db
from yak_server.helpers.logging_helpers import (
    logged_in_successfully,
    modify_password_successfully,
    signed_up_successfully,
)
from yak_server.helpers.password_validator import (
    PasswordRequirements,
    PasswordRequirementsError,
)
from yak_server.helpers.settings import Settings, get_settings
from yak_server.v1.helpers.auth import get_admin_user, get_current_user
from yak_server.v1.helpers.errors import (
    InvalidCredentials,
    NameAlreadyExists,
    UnsatisfiedPasswordRequirements,
    UserNotFound,
)
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.users import (
    CurrentUserOut,
    LoginIn,
    LoginOut,
    ModifyUserIn,
    PasswordRequirementsOut,
    SignupIn,
    SignupOut,
)

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/users", tags=["users"])


@router.post("/signup", status_code=status.HTTP_201_CREATED)
def signup(
    signup_in: SignupIn,
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> GenericOut[SignupOut]:
    try:
        user = signup_user(
            db, signup_in.name, signup_in.first_name, signup_in.last_name, signup_in.password
        )
    except PasswordRequirementsError as password_requirements_error:
        raise UnsatisfiedPasswordRequirements(
            str(password_requirements_error)
        ) from password_requirements_error
    except NameAlreadyExistsError as name_already_exists_error:
        raise NameAlreadyExists(signup_in.name) from name_already_exists_error

    logger.info(signed_up_successfully(user.name))

    return GenericOut(
        result=SignupOut(
            id=user.id,
            name=user.name,
            token=encode_bearer_token(
                sub=user.id,
                expiration_time=pendulum.duration(seconds=settings.jwt_expiration_time),
                secret_key=settings.jwt_secret_key,
            ),
        ),
    )


@router.get("/signup/password_requirements")
def password_requirements() -> GenericOut[PasswordRequirementsOut]:
    password_requirements = PasswordRequirements()

    return GenericOut(
        result=PasswordRequirementsOut(
            minimum_length=password_requirements.MINIMUM_LENGTH,
            uppercase=password_requirements.UPPERCASE,
            lowercase=password_requirements.LOWERCASE,
            digit=password_requirements.DIGIT,
            no_space=password_requirements.NO_SPACE,
        )
    )


@router.post("/login", status_code=status.HTTP_201_CREATED)
def login(
    login_in: LoginIn,
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> GenericOut[LoginOut]:
    user = UserModel.authenticate(db, login_in.name, login_in.password)

    if not user:
        raise InvalidCredentials

    logger.info(logged_in_successfully(user.name))

    return GenericOut(
        result=LoginOut(
            id=user.id,
            name=user.name,
            token=encode_bearer_token(
                sub=user.id,
                expiration_time=pendulum.duration(seconds=settings.jwt_expiration_time),
                secret_key=settings.jwt_secret_key,
            ),
        ),
    )


@router.patch("/{user_id}")
def modify_user(
    user_id: UUID4,
    modify_user_in: ModifyUserIn,
    db: Annotated[Session, Depends(get_db)],
    _: Annotated[UserModel, Depends(get_admin_user)],
) -> GenericOut[CurrentUserOut]:
    user = db.query(UserModel).filter_by(id=user_id).first()
    if not user:
        raise UserNotFound(user_id)

    user.change_password(modify_user_in.password)

    db.commit()

    logger.info(modify_password_successfully(user.name))

    return GenericOut(result=CurrentUserOut.model_validate(user))


@router.get("/current")
def current_user(
    user: Annotated[UserModel, Depends(get_current_user)],
) -> GenericOut[CurrentUserOut]:
    return GenericOut(result=CurrentUserOut.model_validate(user))



import json
from typing import Annotated

import typer

from yak_server import create_app
from yak_server.database import build_engine

from .database import (
    compute_score_board,
    create_admin,
    create_database,
    delete_database,
    drop_database,
    initialize_database,
    setup_migration,
)
from .database.sync import synchronize_official_results
from .env import init_env

app = typer.Typer()


def make_db_app() -> typer.Typer:
    db_app = typer.Typer()

    @db_app.command()
    def create() -> None:
        """Create all database tables."""
        engine = build_engine()
        create_database(engine)

    @db_app.command()
    def init() -> None:
        """Initialize database."""
        app = create_app()
        engine = build_engine()
        initialize_database(engine, app)

    @db_app.command()
    def drop() -> None:
        """Drop all tables."""
        app = create_app()
        engine = build_engine()
        drop_database(engine, debug=app.debug)

    @db_app.command()
    def delete() -> None:
        """Delete all records."""
        app = create_app()
        engine = build_engine()
        delete_database(engine, debug=app.debug)

    @db_app.command()
    def admin(
        password: str = typer.Option(..., prompt=True, confirmation_prompt=True, hide_input=True),
    ) -> None:
        """Create admin account in database."""
        engine = build_engine()
        create_admin(password, engine)

    @db_app.command()
    def migration(*, short: Annotated[bool, typer.Option("--short", "-s")] = False) -> None:
        """Help to run database migration scripts."""
        setup_migration(short=short)

    @db_app.command()
    def score_board() -> None:
        """Compute score board."""
        engine = build_engine()
        compute_score_board(engine)

    @db_app.command()
    def sync() -> None:
        """Synchronize official results and push them to admin with web
        scraping the world cup wikipedia page"""
        engine = build_engine()
        synchronize_official_results(engine)

    return db_app


def make_env_typer() -> typer.Typer:
    env_typer = typer.Typer()

    @env_typer.command()
    def init(  # noqa: PLR0913
        *,
        debug: bool = typer.Option(False, "--debug", "-d", help="Run in debug mode", prompt=True),  # noqa: FBT003
        host: str = typer.Option("localhost", "--host", "-H", help="Database host", prompt=True),
        user: str = typer.Option("yak", "--user", "-u", help="Database username", prompt=True),
        password: str = typer.Option(..., hide_input=True, help="Database password", prompt=True),
        port: int = typer.Option(5432, "--port", "-p", help="Database port", prompt=True),
        database: str = typer.Option(
            "yak_db", "--database", "-D", help="Database name", prompt=True
        ),
        jwt_expiration: int = typer.Option(
            3600, "--jwt-expiration", "-j", help="JWT expiration time in seconds", prompt=True
        ),
        competition: str = typer.Option(
            ..., "--competition", "-c", help="Competition name", prompt=True
        ),
    ) -> None:
        """Build the env files you need to start the server."""
        init_env(debug, host, user, password, competition, database, jwt_expiration, port)

    return env_typer


@app.command()
def openapi() -> None:
    """Print the openapi.json file."""
    app = create_app()
    typer.echo(json.dumps(app.openapi(), separators=(",", ":")))


app.add_typer(make_db_app(), name="db")
app.add_typer(make_env_typer(), name="env")


if __name__ == "__main__":  # pragma: no cover
    app()



from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_group(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    client = TestClient(app_with_valid_jwt_config)

    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_matches_db"),
    )
    initialize_database(engine_for_test, app_with_valid_jwt_config)

    # Signup one random user
    user_name = get_random_string(6)
    first_name = get_random_string(10)
    last_name = get_random_string(8)
    password = get_random_string(8)

    response_signup = client.post(
        "/api/v2",
        json={
            "query": """
                mutation Root(
                    $userName: String!, $firstName: String!,
                    $lastName: String!, $password: String!
                ) {
                    signupResult(
                        userName: $userName, firstName: $firstName,
                        lastName: $lastName, password: $password
                    ) {
                        __typename
                        ... on UserWithToken {
                            token
                        }
                        ... on UserNameAlreadyExists {
                            message
                        }
                    }
                }
            """,
            "variables": {
                "userName": user_name,
                "firstName": first_name,
                "lastName": last_name,
                "password": password,
            },
        },
    )

    assert response_signup.json()["data"]["signupResult"]["__typename"] == "UserWithToken"

    auth_token = response_signup.json()["data"]["signupResult"]["token"]

    # Success case : Get all groups
    query_all_groups = """
        query {
            allGroupsResult {
                __typename
                ... on Groups {
                    groups {
                        id
                        description
                        code
                        phase {
                            id
                            code
                            description
                        }
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
            }
        }
    """

    response_all_groups = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {auth_token}"},
        json={"query": query_all_groups},
    )

    assert response_all_groups.json()["data"]["allGroupsResult"] == {
        "__typename": "Groups",
        "groups": [
            {
                "id": ANY,
                "description": "Groupe A",
                "code": "A",
                "phase": {
                    "id": ANY,
                    "code": "GROUP",
                    "description": "Phase de groupes",
                },
            },
        ],
    }

    # Error case : authentication error
    response_all_groups_with_invalid_token = client.post(
        "/api/v2",
        headers={"Authorization": "Bearer 5555"},
        json={"query": query_all_groups},
    )

    assert response_all_groups_with_invalid_token.json() == {
        "data": {
            "allGroupsResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }
    query_group_by_code = """
        query Root($code: ID!) {
            groupByCodeResult(code: $code) {
                __typename
                ... on Group {
                    id
                    description
                    phase {
                        description
                    }
                    scoreBets {
                        id
                        team1 {
                            score
                            description
                        }
                        team2 {
                            score
                            description
                        }
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on GroupByCodeNotFound {
                    message
                }
            }
        }
    """

    # Success case : Get group by code
    response_group_by_code = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {auth_token}"},
        json={"query": query_group_by_code, "variables": {"code": "A"}},
    )

    assert response_group_by_code.json()["data"]["groupByCodeResult"] == {
        "__typename": "Group",
        "id": ANY,
        "description": "Groupe A",
        "phase": {"description": "Phase de groupes"},
        "scoreBets": [
            {
                "id": ANY,
                "team1": {"score": None, "description": "Andorre"},
                "team2": {"score": None, "description": "Brésil"},
            },
            {
                "id": ANY,
                "team1": {
                    "score": None,
                    "description": "Burkina Faso",
                },
                "team2": {"score": None, "description": "Guatemala"},
            },
            {
                "id": ANY,
                "team1": {
                    "score": None,
                    "description": "Andorre",
                },
                "team2": {
                    "score": None,
                    "description": "Burkina Faso",
                },
            },
            {
                "id": ANY,
                "team1": {
                    "score": None,
                    "description": "Brésil",
                },
                "team2": {
                    "score": None,
                    "description": "Guatemala",
                },
            },
            {
                "id": ANY,
                "team1": {
                    "score": None,
                    "description": "Andorre",
                },
                "team2": {
                    "score": None,
                    "description": "Guatemala",
                },
            },
            {
                "id": ANY,
                "team1": {
                    "score": None,
                    "description": "Brésil",
                },
                "team2": {
                    "score": None,
                    "description": "Burkina Faso",
                },
            },
        ],
    }

    group_id = response_group_by_code.json()["data"]["groupByCodeResult"]["id"]

    # Error case : authentication error
    response_group_by_code_with_invalid_token = client.post(
        "/api/v2", json={"query": query_group_by_code, "variables": {"code": "A"}}
    )

    assert response_group_by_code_with_invalid_token.json() == {
        "data": {
            "groupByCodeResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }

    # Error case : check invalid code
    invalid_group_code = "B"

    response_group_with_invalid_code = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {auth_token}"},
        json={"query": query_group_by_code, "variables": {"code": invalid_group_code}},
    )

    assert response_group_with_invalid_code.json() == {
        "data": {
            "groupByCodeResult": {
                "__typename": "GroupByCodeNotFound",
                "message": f"Group not found: {invalid_group_code}",
            },
        },
    }

    query_by_id = """
        query Root($id: UUID!) {
            groupByIdResult(id: $id) {
                __typename
                ... on Group {
                    description
                    phase {
                        description
                    }
                }
                ... on InvalidToken {
                    message
                }
                ... on ExpiredToken {
                    message
                }
                ... on GroupByIdNotFound {
                    message
                }
            }
        }
    """

    # Success case : Get group by id
    response_group_by_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {auth_token}"},
        json={"query": query_by_id, "variables": {"id": group_id}},
    )

    assert response_group_by_id.json() == {
        "data": {
            "groupByIdResult": {
                "__typename": "Group",
                "description": "Groupe A",
                "phase": {"description": "Phase de groupes"},
            },
        },
    }

    # Error case : Get group with invalid id
    invalid_group_id = str(uuid4())

    response_group_with_invalid_id = client.post(
        "/api/v2",
        headers={"Authorization": f"Bearer {auth_token}"},
        json={"query": query_by_id, "variables": {"id": invalid_group_id}},
    )

    assert response_group_with_invalid_id.json() == {
        "data": {
            "groupByIdResult": {
                "__typename": "GroupByIdNotFound",
                "message": f"Group not found: {invalid_group_id}",
            },
        },
    }

    # Error case : authentication error
    response_group_by_id_with_invalid_token = client.post(
        "/api/v2",
        headers={"Authorization": "Bearer 5555"},
        json={"query": query_by_id, "variables": {"id": group_id}},
    )

    assert response_group_by_id_with_invalid_token.json() == {
        "data": {
            "groupByIdResult": {
                "__typename": "InvalidToken",
                "message": "Invalid token, authentication required",
            }
        }
    }



import sys
from typing import TYPE_CHECKING, Optional

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import pendulum

from .util import get_resources_path

if TYPE_CHECKING:
    from yak_server.helpers.rules import Rules


class MockSettings:
    def __init__(  # noqa: PLR0913
        self,
        *,
        jwt_secret_key: Optional[str] = None,
        jwt_expiration_time: Optional[int] = None,
        data_folder_relative: Optional[str] = None,
        lock_datetime_shift: Optional[pendulum.Duration] = None,
        rules: Optional["Rules"] = None,
        base_correct_result: Optional[int] = None,
        multiplying_factor_correct_result: Optional[int] = None,
        base_correct_score: Optional[int] = None,
        multiplying_factor_correct_score: Optional[int] = None,
        team_qualified: Optional[int] = None,
        first_team_qualified: Optional[int] = None,
        official_results_url: Optional[str] = None,
    ) -> None:
        self.jwt_secret_key = jwt_secret_key
        self.jwt_expiration_time = jwt_expiration_time

        self.data_folder = (
            get_resources_path(data_folder_relative) if data_folder_relative is not None else None
        )

        self.lock_datetime = (
            pendulum.now() + lock_datetime_shift if lock_datetime_shift is not None else None
        )

        self.rules = rules
        self.base_correct_result = base_correct_result
        self.multiplying_factor_correct_result = multiplying_factor_correct_result
        self.base_correct_score = base_correct_score
        self.multiplying_factor_correct_score = multiplying_factor_correct_score
        self.team_qualified = team_qualified
        self.first_team_qualified = first_team_qualified
        self.official_results_url = official_results_url

    def set_lock_datetime(self, lock_datetime_shift: pendulum.Duration) -> None:
        self.lock_datetime = pendulum.now() + lock_datetime_shift

    def __call__(self) -> Self:
        return self



from typing import TYPE_CHECKING, Union

from fastapi import Request
from jwt import ExpiredSignatureError, PyJWTError

from yak_server.database.models import UserModel
from yak_server.helpers.authentication import decode_bearer_token

from .result import ExpiredToken, InvalidToken, UnauthorizedAccessToAdminAPI

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from yak_server.v2.context import YakContext


NUMBER_ELEMENTS_IN_AUTHORIZATION = 2


def _authentify(
    authorization_header: str, db: "Session", jwt_secret_key: str
) -> Union[UserModel, ExpiredToken, InvalidToken]:
    auth_headers = authorization_header.split()

    if len(auth_headers) != NUMBER_ELEMENTS_IN_AUTHORIZATION or auth_headers[0] != "Bearer":
        return InvalidToken()

    token = auth_headers[1]
    try:
        data = decode_bearer_token(token, jwt_secret_key)
    except ExpiredSignatureError:
        return ExpiredToken()
    except PyJWTError:
        return InvalidToken()

    user = db.query(UserModel).filter_by(id=data["sub"]).first()

    if user is None:
        return InvalidToken()

    return user


def authentify(context: "YakContext") -> Union[UserModel, ExpiredToken, InvalidToken]:
    if not isinstance(context.request, Request):  # pragma: no cover
        return InvalidToken()

    return _authentify(
        context.request.headers.get("Authorization", ""),
        context.db,
        context.settings.jwt_secret_key,
    )


def _authentify_admin(
    authorization_header: str,
    db: "Session",
    jwt_secret_key: str,
) -> Union[UserModel, ExpiredToken, InvalidToken, UnauthorizedAccessToAdminAPI]:
    result = _authentify(authorization_header, db, jwt_secret_key)

    if not isinstance(result, UserModel):
        return result

    user = result

    if user.name != "admin":
        return UnauthorizedAccessToAdminAPI()

    return user


def authentify_admin(
    context: "YakContext",
) -> Union[UserModel, ExpiredToken, InvalidToken, UnauthorizedAccessToAdminAPI]:
    if not isinstance(context.request, Request):  # pragma: no cover
        return InvalidToken()

    return _authentify_admin(
        context.request.headers.get("Authorization", ""),
        context.db,
        context.settings.jwt_secret_key,
    )



from collections.abc import Iterable
from typing import TYPE_CHECKING
from uuid import UUID

import strawberry
from sqlalchemy import and_
from strawberry.types import Info

from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)
from yak_server.helpers.group_position import compute_group_rank
from yak_server.helpers.password_validator import PasswordRequirements

from .bearer_authentication import authentify
from .context import YakContext
from .result import (
    AllGroupsResult,
    AllPhasesResult,
    AllTeamsResult,
    AllTeamsSuccessful,
    BinaryBetNotFound,
    BinaryBetResult,
    CurrentUserResult,
    GroupByCodeNotFound,
    GroupByCodeResult,
    GroupByIdNotFound,
    GroupByIdResult,
    GroupRank,
    GroupRankByCodeResult,
    GroupRankByIdResult,
    Groups,
    PhaseByCodeNotFound,
    PhaseByCodeResult,
    PhaseByIdNotFound,
    PhaseByIdResult,
    Phases,
    ScoreBetNotFound,
    ScoreBetResult,
    ScoreBoard,
    ScoreBoardResult,
    TeamByCodeNotFound,
    TeamByCodeResult,
    TeamByIdNotFound,
    TeamByIdResult,
)
from .schema import (
    BinaryBet,
    Group,
    PasswordRequirementsResult,
    Phase,
    ScoreBet,
    Team,
    User,
    UserWithoutSensitiveInfo,
    send_group_position,
)

if TYPE_CHECKING:
    import pendulum
    from sqlalchemy.orm import Session


@strawberry.type
class Query:
    @strawberry.field
    def current_user_result(self, info: Info[YakContext, None]) -> CurrentUserResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        user = result

        db = info.context.db
        settings = info.context.settings

        return User.from_instance(user, db=db, lock_datetime=settings.lock_datetime)

    @strawberry.field
    def password_requirements(self) -> PasswordRequirementsResult:
        password_requirements = PasswordRequirements()

        return PasswordRequirementsResult(
            minimum_length=password_requirements.MINIMUM_LENGTH,
            uppercase=password_requirements.UPPERCASE,
            lowercase=password_requirements.LOWERCASE,
            digit=password_requirements.DIGIT,
            no_space=password_requirements.NO_SPACE,
        )

    @strawberry.field
    def all_teams_result(self, info: Info[YakContext, None]) -> AllTeamsResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db

        return AllTeamsSuccessful(
            teams=[Team.from_instance(team) for team in db.query(TeamModel).all()]
        )

    @strawberry.field
    def team_by_id_result(self, id: UUID, info: Info[YakContext, None]) -> TeamByIdResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db

        team_record = db.query(TeamModel).filter_by(id=id).first()

        if not team_record:
            return TeamByIdNotFound(id=id)

        return Team.from_instance(team_record)

    @strawberry.field
    def team_by_code_result(self, code: str, info: Info[YakContext, None]) -> TeamByCodeResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db

        team_record = db.query(TeamModel).filter_by(code=code).first()

        if not team_record:
            return TeamByCodeNotFound(code=code)

        return Team.from_instance(team_record)

    @strawberry.field
    def score_bet_result(self, id: UUID, info: Info[YakContext, None]) -> ScoreBetResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        score_bet_record = (
            db.query(ScoreBetModel)
            .join(ScoreBetModel.match)
            .filter(and_(ScoreBetModel.id == id, MatchModel.user_id == user.id))
            .first()
        )

        if not score_bet_record:
            return ScoreBetNotFound(id=id)

        return ScoreBet.from_instance(score_bet_record, db=db, lock_datetime=settings.lock_datetime)

    @strawberry.field
    def binary_bet_result(self, id: UUID, info: Info[YakContext, None]) -> BinaryBetResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        binary_bet_record = (
            db.query(BinaryBetModel)
            .join(BinaryBetModel.match)
            .filter(and_(BinaryBetModel.id == id, MatchModel.user_id == user.id))
            .first()
        )

        if not binary_bet_record:
            return BinaryBetNotFound(id=id)

        return BinaryBet.from_instance(
            binary_bet_record, db=db, lock_datetime=settings.lock_datetime
        )

    @strawberry.field
    def all_groups_result(self, info: Info[YakContext, None]) -> AllGroupsResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        return Groups(
            groups=[
                Group.from_instance(
                    group,
                    db=db,
                    user=user,
                    lock_datetime=settings.lock_datetime,
                )
                for group in db.query(GroupModel).order_by(GroupModel.index)
            ],
        )

    @strawberry.field
    def group_by_id_result(self, id: UUID, info: Info[YakContext, None]) -> GroupByIdResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        group_record = db.query(GroupModel).filter_by(id=id).first()

        if not group_record:
            return GroupByIdNotFound(id=id)

        return Group.from_instance(
            group_record,
            db=db,
            user=user,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    def group_by_code_result(
        self,
        code: strawberry.ID,
        info: Info[YakContext, None],
    ) -> GroupByCodeResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        group_record = db.query(GroupModel).filter_by(code=code).first()

        if not group_record:
            return GroupByCodeNotFound(code=code)

        return Group.from_instance(
            group_record,
            db=db,
            user=user,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    def all_phases_result(self, info: Info[YakContext, None]) -> AllPhasesResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        phases = db.query(PhaseModel).order_by(PhaseModel.index)

        return Phases(
            phases=[
                Phase.from_instance(
                    phase,
                    db=db,
                    user=user,
                    lock_datetime=settings.lock_datetime,
                )
                for phase in phases
            ],
        )

    @strawberry.field
    def phase_by_id_result(self, id: UUID, info: Info[YakContext, None]) -> PhaseByIdResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        phase_record = db.query(PhaseModel).filter_by(id=id).first()

        if not phase_record:
            return PhaseByIdNotFound(id=id)

        return Phase.from_instance(
            phase_record,
            db=db,
            user=user,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    def phase_by_code_result(self, code: str, info: Info[YakContext, None]) -> PhaseByCodeResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        phase_record = db.query(PhaseModel).filter_by(code=code).first()

        if not phase_record:
            return PhaseByCodeNotFound(code=code)

        return Phase.from_instance(
            phase_record,
            db=db,
            user=user,
            lock_datetime=settings.lock_datetime,
        )

    @strawberry.field
    def score_board_result(self, info: Info[YakContext, None]) -> ScoreBoardResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db

        users = (
            db.query(UserModel).filter(UserModel.name != "admin").order_by(UserModel.points.desc())
        )

        return ScoreBoard(
            users=[UserWithoutSensitiveInfo.from_instance(user) for user in users],
        )

    @strawberry.field
    def group_rank_by_code_result(
        self,
        code: str,
        info: Info[YakContext, None],
    ) -> GroupRankByCodeResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        group = db.query(GroupModel).filter_by(code=code).first()

        if not group:
            return GroupByCodeNotFound(code=code)

        group_rank = db.query(GroupPositionModel).filter_by(
            user_id=user.id,
            group_id=group.id,
        )

        def send_response(
            db: "Session",
            user: UserModel,
            group: GroupModel,
            group_rank: Iterable[GroupPositionModel],
            lock_datetime: "pendulum.DateTime",
        ) -> GroupRank:
            return GroupRank(
                group_rank=send_group_position(group_rank),
                group=Group.from_instance(
                    group,
                    db=db,
                    user=user,
                    lock_datetime=lock_datetime,
                ),
            )

        if not any(group_position.need_recomputation for group_position in group_rank):
            return send_response(db, user, group, group_rank, settings.lock_datetime)

        score_bets = (
            db.query(ScoreBetModel)
            .join(ScoreBetModel.match)
            .filter(and_(MatchModel.user_id == user.id, MatchModel.group_id == group.id))
        )
        group_rank_result = compute_group_rank(group_rank, score_bets)
        db.commit()

        return send_response(db, user, group, group_rank_result, settings.lock_datetime)

    @strawberry.field
    def group_rank_by_id_result(
        self,
        id: UUID,
        info: Info[YakContext, None],
    ) -> GroupRankByIdResult:
        result = authentify(info.context)

        if not isinstance(result, UserModel):
            return result

        db = info.context.db
        user = result
        settings = info.context.settings

        group = db.query(GroupModel).filter_by(id=id).first()

        if not group:
            return GroupByIdNotFound(id=id)

        group_rank = db.query(GroupPositionModel).filter_by(
            user_id=user.id,
            group_id=group.id,
        )

        def send_response(
            db: "Session",
            user: UserModel,
            group: GroupModel,
            group_rank: Iterable[GroupPositionModel],
            lock_datetime: "pendulum.DateTime",
        ) -> GroupRank:
            return GroupRank(
                group_rank=send_group_position(group_rank),
                group=Group.from_instance(
                    group,
                    db=db,
                    user=user,
                    lock_datetime=lock_datetime,
                ),
            )

        if not any(group_position.need_recomputation for group_position in group_rank):
            return send_response(db, user, group, group_rank, settings.lock_datetime)

        score_bets = (
            db.query(ScoreBetModel)
            .join(ScoreBetModel.match)
            .filter(and_(MatchModel.user_id == user.id, MatchModel.group_id == group.id))
        )
        group_rank_result = compute_group_rank(group_rank, score_bets)
        db.commit()

        return send_response(db, user, group, group_rank_result, settings.lock_datetime)



from http import HTTPStatus
from secrets import SystemRandom
from typing import TYPE_CHECKING
from unittest.mock import ANY
from uuid import uuid4

import pendulum
from starlette.testclient import TestClient

from testing.mock import MockSettings
from testing.util import get_random_string
from yak_server.cli.database import initialize_database
from yak_server.helpers.settings import get_settings

if TYPE_CHECKING:
    import pytest
    from fastapi import FastAPI
    from sqlalchemy import Engine


def test_create_score_bet(
    app_with_valid_jwt_config: "FastAPI",
    engine_for_test: "Engine",
    monkeypatch: "pytest.MonkeyPatch",
) -> None:
    monkeypatch.setattr(
        "yak_server.cli.database.get_settings",
        MockSettings(data_folder_relative="test_create_bet"),
    )

    initialize_database(engine_for_test, app_with_valid_jwt_config)

    client = TestClient(app_with_valid_jwt_config)

    # Signup one user
    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": get_random_string(6),
            "first_name": get_random_string(6),
            "last_name": get_random_string(6),
            "password": get_random_string(100),
        },
    )

    assert response_signup.status_code == HTTPStatus.CREATED

    token = response_signup.json()["result"]["token"]

    # Fetch all teams
    response_all_teams = client.get("/api/v1/teams")

    assert response_all_teams.status_code == HTTPStatus.OK

    team1 = response_all_teams.json()["result"]["teams"][0]
    team2 = response_all_teams.json()["result"]["teams"][1]
    score1 = SystemRandom().randrange(1, 5)

    # Fetch all groups
    response_all_groups = client.get("/api/v1/groups", headers={"Authorization": f"Bearer {token}"})

    assert response_all_groups.status_code == HTTPStatus.OK

    group_id = response_all_groups.json()["result"]["groups"][0]["id"]

    # Create one score bet
    response_create_score_bet = client.post(
        "/api/v1/score_bets",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "index": 1,
            "team1": {"id": team1["id"], "score": score1},
            "team2": {"id": team2["id"]},
            "group": {"id": group_id},
        },
    )

    score_bet_response = {
        "ok": True,
        "result": {
            "group": {
                "code": "A",
                "description": "Groupe A",
                "id": group_id,
            },
            "phase": {
                "code": "GROUP",
                "description": "Phase de groupes",
                "id": ANY,
            },
            "score_bet": {
                "id": ANY,
                "locked": False,
                "team1": {**team1, "score": score1},
                "team2": {**team2, "score": None},
            },
        },
    }

    assert response_create_score_bet.json() == score_bet_response

    score_bet_id = response_create_score_bet.json()["result"]["score_bet"]["id"]

    # Check GET after POST
    response_retrieve_one_bet = client.get(
        f"/api/v1/score_bets/{score_bet_id}",
        headers={"Authorization": f"Bearer {token}"},
    )

    assert response_retrieve_one_bet.status_code == HTTPStatus.OK
    assert response_retrieve_one_bet.json() == score_bet_response

    # Check POST with invalid group id
    invalid_group_id = str(uuid4())

    response_create_score_bet_with_group_id = client.post(
        "/api/v1/score_bets",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "index": 1,
            "team1": {"id": team1["id"]},
            "team2": {"id": team2["id"]},
            "group": {"id": invalid_group_id},
        },
    )

    assert response_create_score_bet_with_group_id.status_code == HTTPStatus.NOT_FOUND
    assert response_create_score_bet_with_group_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Group not found: {invalid_group_id}",
    }

    # Check POST with invalid team1 id
    invalid_team1_id = str(uuid4())

    response_create_score_bet_with_team1_id = client.post(
        "/api/v1/score_bets",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "index": 1,
            "team1": {"id": invalid_team1_id},
            "team2": {"id": team2["id"]},
            "group": {"id": group_id},
        },
    )

    assert response_create_score_bet_with_team1_id.status_code == HTTPStatus.NOT_FOUND
    assert response_create_score_bet_with_team1_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Team not found: {invalid_team1_id}",
    }

    # Check POST with invalid team2 id
    invalid_team2_id = str(uuid4())

    response_create_score_bet_with_team2_id = client.post(
        "/api/v1/score_bets",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "index": 1,
            "team1": {"id": team1["id"]},
            "team2": {"id": invalid_team2_id},
            "group": {"id": group_id},
        },
    )

    assert response_create_score_bet_with_team2_id.status_code == HTTPStatus.NOT_FOUND
    assert response_create_score_bet_with_team2_id.json() == {
        "ok": False,
        "error_code": HTTPStatus.NOT_FOUND,
        "description": f"Team not found: {invalid_team2_id}",
    }

    # Check bet locking
    app_with_valid_jwt_config.dependency_overrides[get_settings]().set_lock_datetime(
        -pendulum.duration(minutes=10)
    )

    response_create_locked_score_bet = client.post(
        "/api/v1/score_bets",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "index": 1,
            "team1": {"id": team1["id"], "score": score1},
            "team2": {"id": team2["id"]},
            "group": {"id": group_id},
        },
    )

    assert response_create_locked_score_bet.status_code == HTTPStatus.UNAUTHORIZED
    assert response_create_locked_score_bet.json() == {
        "ok": False,
        "error_code": HTTPStatus.UNAUTHORIZED,
        "description": "Cannot modify score bet, lock date is exceeded",
    }



import strawberry
from graphql.validation import NoSchemaIntrospectionCustomRule
from strawberry.extensions import (
    AddValidationRules,
    IgnoreContext,
    MaxAliasesLimiter,
    MaxTokensLimiter,
    ParserCache,
    QueryDepthLimiter,
    ValidationCache,
)

from .mutation import Mutation
from .query import Query


def get_schema(*, debug: bool) -> strawberry.Schema:
    def should_ignore(_: IgnoreContext) -> bool:
        return False

    extensions = [
        QueryDepthLimiter(max_depth=6, should_ignore=should_ignore),
        MaxAliasesLimiter(max_alias_count=30),
        MaxTokensLimiter(max_token_count=500),
        ValidationCache(),
        ParserCache(),
    ]

    if not debug:
        extensions.append(AddValidationRules([NoSchemaIntrospectionCustomRule]))

    return strawberry.Schema(
        query=Query,
        mutation=Mutation,
        extensions=extensions,
    )



from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from sqlalchemy import and_, update

from yak_server.database.models import GroupPositionModel, MatchModel, ScoreBetModel

if TYPE_CHECKING:
    from uuid import UUID

    from sqlalchemy.orm import Session

    from yak_server.database.models import UserModel


def create_group_position(score_bets: Iterable[ScoreBetModel]) -> list[GroupPositionModel]:
    team_ids = []

    group_positions = []

    for score_bet in score_bets:
        if score_bet.match.team1_id is not None and score_bet.match.team1_id not in team_ids:
            group_positions.append(
                GroupPositionModel(
                    team_id=score_bet.match.team1_id,
                    user_id=score_bet.match.user.id,
                    group_id=score_bet.match.group.id,
                ),
            )
            team_ids.append(score_bet.match.team1_id)

        if score_bet.match.team2_id is not None and score_bet.match.team2_id not in team_ids:
            group_positions.append(
                GroupPositionModel(
                    team_id=score_bet.match.team2_id,
                    user_id=score_bet.match.user.id,
                    group_id=score_bet.match.group.id,
                ),
            )
            team_ids.append(score_bet.match.team2_id)

    return group_positions


@dataclass
class GroupPosition:
    won: int = 0
    drawn: int = 0
    lost: int = 0
    goals_for: int = 0
    goals_against: int = 0


def compute_group_rank(
    group_rank: Iterable[GroupPositionModel], score_bets: Iterable["ScoreBetModel"]
) -> list[GroupPositionModel]:
    new_group_position: dict[UUID, GroupPosition] = {}

    for score_bet in score_bets:
        team1_id = score_bet.match.team1_id
        team2_id = score_bet.match.team2_id

        if team1_id not in new_group_position:
            new_group_position[team1_id] = GroupPosition()

        if team2_id not in new_group_position:
            new_group_position[team2_id] = GroupPosition()

        if score_bet.score1 is None or score_bet.score2 is None:
            continue

        new_group_position[team1_id].goals_for += score_bet.score1
        new_group_position[team1_id].goals_against += score_bet.score2
        new_group_position[team2_id].goals_for += score_bet.score2
        new_group_position[team2_id].goals_against += score_bet.score1

        if score_bet.score1 > score_bet.score2:
            new_group_position[team1_id].won += 1
            new_group_position[team2_id].lost += 1
        elif score_bet.score1 == score_bet.score2:
            new_group_position[team1_id].drawn += 1
            new_group_position[team2_id].drawn += 1
        else:
            new_group_position[team1_id].lost += 1
            new_group_position[team2_id].won += 1

    for group_position in group_rank:
        group_position.won = new_group_position[group_position.team_id].won
        group_position.drawn = new_group_position[group_position.team_id].drawn
        group_position.lost = new_group_position[group_position.team_id].lost
        group_position.goals_for = new_group_position[group_position.team_id].goals_for
        group_position.goals_against = new_group_position[group_position.team_id].goals_against
        group_position.need_recomputation = False

    return sorted(
        group_rank,
        key=lambda team_result: (
            team_result.points,
            team_result.goals_difference,
            team_result.goals_for,
        ),
        reverse=True,
    )


def get_group_rank_with_code(
    db: "Session", user: "UserModel", group_id: "UUID"
) -> list[GroupPositionModel]:
    group_rank = db.query(GroupPositionModel).filter_by(group_id=group_id, user_id=user.id)

    if not any(group_position.need_recomputation for group_position in group_rank):
        return sorted(
            group_rank,
            key=lambda team_result: (
                team_result.points,
                team_result.goals_difference,
                team_result.goals_for,
            ),
            reverse=True,
        )

    score_bets = (
        db.query(ScoreBetModel)
        .join(ScoreBetModel.match)
        .filter(and_(MatchModel.user_id == user.id, MatchModel.group_id == group_id))
    )

    group_rank_list = compute_group_rank(group_rank, score_bets)

    db.commit()

    return sorted(
        group_rank_list,
        key=lambda team_result: (
            team_result.points,
            team_result.goals_difference,
            team_result.goals_for,
        ),
        reverse=True,
    )


def set_recomputation_flag(db: "Session", team_id: Optional["UUID"], user_id: "UUID") -> None:
    db.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == team_id,
            GroupPositionModel.user_id == user_id,
            GroupPositionModel.need_recomputation.is_(False),
        ),
    )
